from __future__ import annotations

from typing import Dict
from typing import Any
from typing import List

import os
import sys
import traceback

from thumbny.models import TemplateModel
from thumbny.templates_manager.file_handler import FileHandler
from thumbny.templates_manager.validator import Validator
from thumbny.exceptions import NotAbleToCreate


class TemplateManager:
    def __new__(cls) -> TemplateManager:
        if not hasattr(cls, "_instance"):
            cls._instance = super(TemplateManager, cls).__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self.templates_path = os.path.join(sys.path[0],
                                           "thumny-templates",
                                           "templates")

        self.file_handler = FileHandler(self, self.templates_path)
        self.validtor = Validator(self, self.templates_path)

    def create(self, model: TemplateModel) -> None:
        self.validtor.validate_tempalate_key(model.key)
        self.file_handler.create_template_dir(self.templates_path)
        template_Path = self.file_handler.create_template_structure(model.key)

        try:
            self.file_handler.copy_assets(model, template_Path)
            self.file_handler.save_config(model, template_Path)
        except Exception:
            self.file_handler.delete_template(model.key)
            raise NotAbleToCreate(traceback.format_exc())

    def delete(self, name: str) -> None:
        self.file_handler.delete_template(name)

    def get_all_templates(self) -> List[str]:
        return self.file_handler.get_all_templates()

    def get_template_info(self, name: str) -> Dict[str, Any]:
        return self.file_handler.get_template_info(name)



from thumbny.commands.create import CreateCommand
from thumbny.commands.delete import DeleteCommand
from thumbny.commands.generate import GenerateCommand
from thumbny.commands.templates import TemplatesCommand
from thumbny.commands.info import InfoCommand

__all__ = [
    "CreateCommand",
    "DeleteCommand",
    "GenerateCommand",
    "InfoCommand",
    "TemplatesCommand",
]



from typing import get_type_hints
from typing import get_origin
from typing import get_args
from typing import Union
from typing import Any

import re
from dataclasses import dataclass


HEX_REGEX = r'^#[0-9a-fA-F]{6}$'


def _is_optional(field_type) -> bool:
    origin = get_origin(field_type)
    if origin is Union:
        return type(None) in get_args(field_type)
    return False


def check_required_fields(instance: dataclass) -> None:
    annotations = get_type_hints(instance.__class__)
    for field_name, field_type in annotations.items():
        if (not _is_optional(field_type)
                and getattr(instance, field_name) is None):
            raise ValueError(f"Field '{field_name}' cannot be None.")


def check_spaces(field_name: str, field_value: Any) -> None:
    if field_value.count(" "):
        raise ValueError(f"{field_name} cannot contain spaces")


def check_hex_color(field_name: str, field_value: Any) -> None:
    if not re.match(HEX_REGEX, str(field_value)):
        raise ValueError(f"{field_name} isn't correct. "
                         "Should follow this format: #FFFFFF "
                         f"Given ({field_value})")



from typing import Optional

from thumbny.base import RunnerBase
from thumbny.commands import TemplatesCommand


class TemplatesRunner(RunnerBase):
    def __init__(self, json_string: Optional[str] = None) -> None:
        super().__init__(json_string)

    def execute(self) -> None:
        command = TemplatesCommand()
        return command.execute()



from enum import Enum


class PositionTypeEnum(Enum):
    RELATIVE = "relative"
    ABSOLUTE = "absolute"


class XPositionEnum(Enum):
    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


class YPositionEnum(Enum):
    TOP = "top"
    CENTER = "center"
    BOTTOM = "bottom"



from __future__ import annotations

from typing import Dict
from typing import List
from typing import Optional
from typing import Any
from dataclasses import dataclass

from thumbny.models.shared import TagModel
from thumbny.models.validation import check_required_fields
from thumbny.models.validation import check_spaces
from thumbny.models.validation import check_hex_color


@dataclass
class PaddingModel:
    top: int
    bottom: int
    left: int
    right: int

    def __post_init__(self) -> None:
        check_required_fields(self)

    @classmethod
    def make(cls, data: Dict[str, Any]) -> PaddingModel:
        return cls(top=data.get("top"),
                   bottom=data.get("bottom"),
                   left=data.get("left"),
                   right=data.get("right"))


@dataclass
class LabelModel:
    key: str
    content: str
    position: TagModel
    padding: Optional[PaddingModel]
    font_color: str
    font_size: int
    font_family: Optional[str]
    alignment: Optional[str]

    def __post_init__(self) -> None:
        check_required_fields(self)

    @classmethod
    def make(cls, data: Dict[str, Any]) -> LabelModel:
        position = TagModel.make(data.get("position"))

        padding = None
        if data.get("padding"):
            padding = PaddingModel.make(data.get("padding"))

        return cls(key=data.get("key"),
                   content=data.get("content"),
                   position=position,
                   padding=padding,
                   font_color=data.get("font_color"),
                   font_size=data.get("font_size"),
                   font_family=data.get("font_family"),
                   alignment=data.get("alignment"),)


@dataclass
class TemplateModel:
    key: str
    name: str
    width: int
    height: int
    background_color: str
    background_image: Optional[str]
    labels: List[LabelModel]

    def __post_init__(self) -> None:
        check_required_fields(self)
        check_spaces("key", self.key)
        check_hex_color("background_color", self.background_color)

    @classmethod
    def make(cls, data: Dict[str, Any]) -> TemplateModel:
        labels = []
        for label in data.get("labels", []):
            label = LabelModel.make(label)
            labels.append(label)

        return cls(key=data.get("key"),
                   name=data.get("name"),
                   width=data.get("width"),
                   height=data.get("height"),
                   background_color=data.get("background_color"),
                   background_image=data.get("background_image"),
                   labels=labels)



from typing import Optional

from thumbny.base import RunnerBase
from thumbny.commands import GenerateCommand
from thumbny.models import FillerModel


class GenerateRunner(RunnerBase):
    def __init__(self, json_string: Optional[str] = None) -> None:
        super().__init__(json_string)

    def build(self, json_dict: dict) -> FillerModel:
        return FillerModel.make(json_dict)

    def execute(self) -> None:
        command = GenerateCommand(self.model)
        command.execute()



from thumbny.templates_manager import TemplateManager


class CommandBase:
    def __init__(self) -> None:
        self.tm = TemplateManager()

    def execute() -> None:
        raise NotImplementedError()



import argparse
from argparse import ArgumentParser as AP
from argparse import _SubParsersAction
from argparse import Namespace

from thumbny.base import runner
from thumbny.arguments_parser.create import CreateRunner
from thumbny.arguments_parser.delete import DeleteRunner
from thumbny.arguments_parser.generate import GenerateRunner
from thumbny.arguments_parser.info import InfoRunner
from thumbny.arguments_parser.templates import TemplatesRunner


COMMANDS = {"create": CreateRunner,
            "delete": DeleteRunner,
            "generate": GenerateRunner,
            "info": InfoRunner,
            "templates": TemplatesRunner,
            }


class Parser:
    def _register_create(self, subparser: _SubParsersAction) -> None:
        parser: AP = subparser.add_parser("create",
                                          help="Create a new template")

        parser.add_argument("-d", "--data",
                            help="json data")

    def _register_delete(self, subparser: _SubParsersAction) -> None:
        parser: AP = subparser.add_parser("delete",
                                          help="Delete a template")

        parser.add_argument("-d", "--data",
                            help="json data")

    def _register_generate(self, subparser: _SubParsersAction) -> None:
        parser: AP = subparser.add_parser("generate",
                                          help="Generate a thumbnail")

        parser.add_argument("-d", "--data",
                            help="json data")

    def _register_info(self, subparser: _SubParsersAction) -> None:
        parser: AP = subparser.add_parser("info", help="Info of a template")

        parser.add_argument("-d", "--data",
                            help="json data")

    def _register_templates(self, subparser: _SubParsersAction) -> None:
        subparser.add_parser("templates", help="List all templates")

    def _execute(self, args: Namespace) -> None:
        command_ref = COMMANDS.get(args.command)
        arguments = dict(args._get_kwargs())
        command: runner = command_ref(json_string=arguments.get("data"))
        command.execute()

    def parse(self) -> None:
        parser = argparse.ArgumentParser()
        subparser = parser.add_subparsers(dest="command", required=True)
        self._register_create(subparser)
        self._register_delete(subparser)
        self._register_generate(subparser)
        self._register_info(subparser)
        self._register_templates(subparser)
        args = parser.parse_args()
        self._execute(args)





class TemplateExist(Exception):
    ...


class TemplateNotExist(Exception):
    ...


class FontNotFound(Exception):
    ...


class NotValidFontExtension(Exception):
    ...


class NotAbleToCreate(Exception):
    ...
